@@ -247,7 +247,16 @@ class SMTPConnection extends EventEmitter {
                 .join(' ');
         }
 
-        if (this._socket && this._socket.writable) {
+        if (code >= 400) {
+            this.session.error = payload;
+        }
+
+        // Ref. https://datatracker.ietf.org/doc/html/rfc4954#section-4
+        if (code === 334 && payload === '334') {
+            payload += ' ';
+        }
+
+        if (this._socket && !this._socket.destroyed && this._socket.readyState === 'open') {
             this._socket.write(payload + '\r\n');
             this._server.logger.debug(
                 {
@@ -284,7 +293,7 @@ class SMTPConnection extends EventEmitter {
      * Setup socket event handlers
      */
     _setListeners(callback) {
-        this._socket.on('close', () => this._onClose());
+        this._socket.on('close', hadError => this._onCloseEvent(hadError));
         this._socket.on('error', err => this._onError(err));
         this._socket.setTimeout(this._server.options.socketTimeout || SOCKET_TIMEOUT, () => this._onTimeout());
         this._socket.pipe(this._parser);
@@ -294,13 +303,30 @@ class SMTPConnection extends EventEmitter {
         this.upgrade(() => false, callback);
     }
 
+    _onCloseEvent(hadError) {
+        this._server.logger.info(
+            {
+                tnx: 'close',
+                cid: this.id,
+                host: this.remoteAddress,
+                user: (this.session.user && this.session.user.username) || this.session.user,
+                hadError
+            },
+            '%s received "close" event from %s' + (hadError ? ' after error' : ''),
+            this.id,
+            this.remoteAddress
+        );
+
+        this._onClose();
+    }
+
     /**
      * Fired when the socket is closed
      * @event
      */
     _onClose(/* hadError */) {
         if (this._parser) {
-            this._parser.closed = true;
+            this._parser.isClosed = true;
             this._socket.unpipe(this._parser);
             this._parser = false;
         }
@@ -339,6 +365,19 @@ class SMTPConnection extends EventEmitter {
      * @param {Error} err Error object
      */
     _onError(err) {
+        err.remote = this.remoteAddress;
+        this._server.logger.error(
+            {
+                err,
+                tnx: 'error',
+                user: (this.session.user && this.session.user.username) || this.session.user
+            },
+            '%s %s %s',
+            this.id,
+            this.remoteAddress,
+            err.message
+        );
+
         if ((err.code === 'ECONNRESET' || err.code === 'EPIPE') && (!this.session.envelope || !this.session.envelope.mailFrom)) {
             // We got a connection error outside transaction. In most cases it means dirty
             // connection ending by the other party, so we can just ignore it
@@ -346,15 +385,6 @@ class SMTPConnection extends EventEmitter {
             return;
         }
 
-        this._server.logger.error(
-            {
-                err,
-                tnx: 'error',
-                user: (this.session.user && this.session.user.username) || this.session.user
-            },
-            '%s',
-            err.message
-        );
         this.emit('error', err);
     }
 
@@ -374,11 +404,7 @@ class SMTPConnection extends EventEmitter {
      * @param {Function} callback Callback to run once the command is processed
      */
     _onCommand(command, callback) {
-        let commandName = (command || '')
-            .toString()
-            .split(' ')
-            .shift()
-            .toUpperCase();
+        let commandName = (command || '').toString().split(' ').shift().toUpperCase();
         this._server.logger.debug(
             {
                 tnx: 'command',
@@ -478,10 +504,7 @@ class SMTPConnection extends EventEmitter {
      * @returns {Boolean} Returns true if the command can be used
      */
     _isSupported(command) {
-        command = (command || '')
-            .toString()
-            .trim()
-            .toUpperCase();
+        command = (command || '').toString().trim().toUpperCase();
         return !this._server.options.disabledCommands.includes(command) && typeof this['handler_' + command] === 'function';
     }
 
@@ -494,20 +517,11 @@ class SMTPConnection extends EventEmitter {
      */
     _parseAddressCommand(name, command) {
         command = (command || '').toString();
-        name = (name || '')
-            .toString()
-            .trim()
-            .toUpperCase();
+        name = (name || '').toString().trim().toUpperCase();
 
         let parts = command.split(':');
-        command = parts
-            .shift()
-            .trim()
-            .toUpperCase();
-        parts = parts
-            .join(':')
-            .trim()
-            .split(/\s+/);
+        command = parts.shift().trim().toUpperCase();
+        parts = parts.join(':').trim().split(/\s+/);
 
         let address = parts.shift();
         let args = false;
@@ -648,10 +662,7 @@ class SMTPConnection extends EventEmitter {
      * Processes EHLO. Requires valid hostname as the single argument.
      */
     handler_EHLO(command, callback) {
-        let parts = command
-            .toString()
-            .trim()
-            .split(/\s+/);
+        let parts = command.toString().trim().split(/\s+/);
         let hostname = parts[1] || '';
 
         if (parts.length !== 2) {
@@ -695,10 +706,7 @@ class SMTPConnection extends EventEmitter {
      * Processes HELO. Requires valid hostname as the single argument.
      */
     handler_HELO(command, callback) {
-        let parts = command
-            .toString()
-            .trim()
-            .split(/\s+/);
+        let parts = command.toString().trim().split(/\s+/);
         let hostname = parts[1] || '';
 
         if (parts.length !== 2) {
@@ -777,10 +785,7 @@ class SMTPConnection extends EventEmitter {
         }
 
         let allowedKeys = ['NAME', 'ADDR', 'PORT', 'PROTO', 'HELO', 'LOGIN'];
-        let parts = command
-            .toString()
-            .trim()
-            .split(/\s+/);
+        let parts = command.toString().trim().split(/\s+/);
         let key, value;
         let data = new Map();
         parts.shift(); // remove XCLIENT prefix
@@ -972,10 +977,7 @@ class SMTPConnection extends EventEmitter {
         }
 
         let allowedKeys = ['NAME', 'ADDR', 'PORT', 'PROTO', 'HELO', 'IDENT', 'SOURCE'];
-        let parts = command
-            .toString()
-            .trim()
-            .split(/\s+/);
+        let parts = command.toString().trim().split(/\s+/);
         let key, value;
         let data = new Map();
         let hasAddr = false;
@@ -1124,10 +1126,7 @@ class SMTPConnection extends EventEmitter {
      * Check if selected authentication is available and delegate auth data to SASL
      */
     handler_AUTH(command, callback) {
-        let args = command
-            .toString()
-            .trim()
-            .split(/\s+/);
+        let args = command.toString().trim().split(/\s+/);
         let method;
         let handler;
 
@@ -1325,10 +1324,7 @@ class SMTPConnection extends EventEmitter {
      * Processes sendmail WIZ command, upgrades to "wizard mode"
      */
     handler_WIZ(command, callback) {
-        let args = command
-            .toString()
-            .trim()
-            .split(/\s+/);
+        let args = command.toString().trim().split(/\s+/);
         let password;
 
         args.shift(); // remove WIZ
@@ -1414,7 +1410,7 @@ class SMTPConnection extends EventEmitter {
         // upgrade connection
         let secureSocket = new tls.TLSSocket(this._socket, socketOptions);
 
-        secureSocket.once('close', () => this._onClose());
+        secureSocket.once('close', hadError => this._onCloseEvent(hadError));
         secureSocket.once('error', err => this._onError(err));
         secureSocket.once('_tlsError', err => this._onError(err));
         secureSocket.once('clientError', err => this._onError(err));
